import pytest

from ldf_adapter import UserInfo, User

import settings


@pytest.mark.parametrize("data", settings.ALL_INPUT)
def test_get_status(user):
    # assert user.assurance_verifier() == True
    assert user.get_status().attributes == {"state": "not_deployed", "message": "No message"}




import pytest

from itertools import repeat

from ldf_adapter import UserInfo, CONFIG
import settings


@pytest.mark.parametrize('sub', [
        "MWMQb4ybpHVSThMGpRKkqFDJIYlGLXl1CWXSRgM8bQGR9mMXRXtMbLFubL8Sua6vZn8Dq9X3YGoKR",
        "bNAkgXeaN2rlP83UeckV0fSjU2qNmKjQ7BsOsGFC7KB1PHtYGxRXkdSZ6S1egB085cwkIYt0NNPe",
        "0Y7zkfbvLmFVclSBFqtioE0xAaV4ZtMJ7tHxScwAN6FKoPn9R3aSXjoqp1jxRWFnyN7kNoq0nvJ33f7",
        "CO6bVYGIPtzPMHpm9o13S1bztxv6jHVEAXsGX3yBRmSv8RNnVsyzjJ67bvuvl4Tq2T8rPNTcfRRqgFF",
        "X5MOmSBpHzNj3h2BzHwhCFnHzMEpT7BxJEh6X0sgpqp7TBuUdLnfBIoovpQGEYZ7zxloVgmBV5B0FUa",
        "QHp9whOAuvUgmbJBUas1haLQfa4y4VnRlxbJwqm7pqvOfZudZAbDCXlIi20ifqfBYyjQpFN9XbOUSHi",
        "bQZO9gf6K08rlRilSYNlIuroaUBvlTV2QmHFHUlm8HUWZHs8RlNnvxQZXZZ62Oju3GARJQTx8Vb8vkF",
        "YqZE73sSxnOgoJ4jW0AhFfyZZPoaQGuzWZljOL4Nmt4x6VITfrH8EzJNES2uc0mOfGZrEIWUrxCHD6J",
        "LrcvjVB6q3omCRjzoX2uLJZzurDcU0srxJqoSmOSdCvdxisxvRo9nj52jCUxYCEnEW3FV4fKFt25Cqq4",
        "p32UOxqQGTh7xD8mfY1w9XVc0j2II06oXjw19iGhWUDMOMJiOFbwbwQRR6XVyHDwnR5hPA4bWlAxi6SK",
        "TRJvjHpnkwdsKpCYyjOPbbk5Zvk5YDfaO0BcRXpyntpuUQg5WLu9esrkqxHxH8o28A6An2bSG8mIP",
        "wl1dBQBFgPHFgetwz9m6VOYeTzGn2Pu3AMDtPRX6V0XvzhxhL7y9iEatlPqiSl5OO0yiSQWDz9GJ4Ut",
        "wTAg6ewnKtcpdj1t83NWojg7dr7ydjMhZf93hLCR5NA4GtlVfAJGwmy5XGO0AVMWMO0jhAw2KMhVK5t",
        "IkYQVi17paCZdkURaFput1p5FA2rCemT8szPIwtlDnbeecVuzEdsYVT7C3zXL8TTPKQXDdMxKYbibWO5",
        "yp1Xe4fhSGwdEyqzeejRwL8wxVeiMmamjcydJmR5hjIbQ5AQoN7pe4PwGSvNKXvkQmUO09ju0gVFALh",
        "QP8b8AwjfCw7n20WAyOYDCLztANZaKvc4uCyVtYwJnFWHYaXqFLlolPc9LJ7bb7shaRvFDpQHt07Qo",
        "nScpPJSkxQja7eMtgjM83MILbYarScUR4PWlLXK40rLHavl4dDty2OV6QJLH8AVs7LMtaOrWieHAiY",
        "UuTbVNi2RknFwKfy2n3XfCqkrscrIi0cdpJXVrbzIxBqgcvOOCWGL9YFXVTjWhbFjCvzxFgTR22yKw0X",
        "Vb15x73G3eYqtCEaqjM2MaGtUmKxNansoJxPhff5wtIK1VhDO6PutZrYxEAsxoZYIrhIZMrZpc926h",
        "w72mv9VQ62SDcBGFw4Izw7Vj5eH0JYRXdN0xImmIfZtAszqFz5mrB5aBxTZKqieTExDJWlrlAGfHIAZ"
    ])
def test_sub_masked_for_bwidm_eppn_unchanged(sub):
    info = UserInfo({'user': {'userinfo': {'sub': sub}}})
    assert info._sub_masked_for_bwidm_eppn() == sub


# Generated with: cat /dev/urandom | tr -dc '[:graph:]' | tr -d 'a-zA-Z0-9_!#$%&*+/=?{|}~^.\-' | fold -w 80 | head -n 20 | sed 's/\\/\\\\/g;s/'\''/\\'\''/g;s/^/'\''/;s/$/'\'',/'
@pytest.mark.parametrize('sub', [
        ')@\\\'>""\'`,,();@"",,[];>,@,((::;;\'\']<)[:]"@;<\\<",])]>\'][@`<[\\<)>();("`"@\';;[(;[@;',
        ',`@:(,(>>;,()],["(@,\'"<]:[[,::;,@>`<[;\\:,@,(",:;)]`::;;>@<](><:(],@]\\:\\\'\'>>:\\\\>"',
        '@;<@,)]`[],:,,;\\>,](":,]@"@\'>)>]]@(]```(\\(),"(\\[,"(<`<[):;>`(:`,>,\':<[>`";>),(@@',
        '<`:;>``,@";))[\']@((>:)>`@\'":)>:"`,)><,\\""`,\';)<`\\>]`\'((\\>("<]\\\\\'>\\>\\\\"">],\\[<\'[\\',
        '\'\'\\\\\'>><\'>]\'`\'<)>,(;:,`)\'>\'];,,]@@";]]\',[\\([;`([>@[\\:;)@(<[@"`,`@\'\'"));,\\><`>>,@',
        '(`>@]`)\':]>@""@@<:;::,>\\\\(":`<`]\'(<@\')[\\\'(\'@\\:";"`(<[:][[@<,"),<\\,;));\'>`\'\\>\':[)',
        '(]`\\"\\`\'(;;)\\[\\)`\\[`)](>`]`)\\[]\\:>::@;\\[\',[><`:">@@,(@\'`[@@````,`<,\'])(\\[]\'(]<[<',
        ']<[\'`)<`[)>\\<]\']@>,(""["<,":](""\'>,,\'@\\]><)\\\',(\\"("[``\'@];<"(:>[>\')\\;)::,[:;>",[',
        '"(()\'\'\\\'@)]:;:;\'<":])`\\(`\'(](),@;:]\'>);";\'\\((::[;><)\'(<[`(::(>>(@)(;\\,@[`]):,`,)',
        '`)<"\']@:@):"]`\\);:\\\\`">)\\@>><;>)],<:;\\],`<[]])(\'\\\\,(,\\:`(@>;\'[))<,,"];@[`,<>\\\\\';',
        '(>`";>\'\'@;)")\'\\;<\\`<:<,"]@:(:)">\\@\\<[;)[:<@"\\(],(":)<\'"<,<`>:)@\';@`]:,[\\\\@<[\'>>\\',
        '])``\\;"]`):\\`),):\';\';[\'")>,`[<">":>>(\'](,[)<\'@;:[[@@]`@;((@<<;`\\():;[:<,\'`>>\\[",',
        ']]"<@``;;<<]);<];):\'[,<<>\\>))>(>`)")\'(`<]>:@<:;,@>)\';(:)>))\'\';::>[]<]\'`)@["<`<](',
        ')<>][`";["",():`]@@[`];(;)\'\\\':,]`[\\[@]\'"";":<",[)())\\<,];"<,\'"<,:]"::(]<>@](@)<\\',
        '<;::>\'];(>[;;],]:(@<<:",<\'>\\@,""`(@\\\\@)@\'<[`,:<(\'`;<@\\"><@>:[](,`<]""`@"[""[)(\\:',
        '::>"(:(())[\\\'(,[\'((`):\']\';`"]`@,\';:];@<`<:\\,;:">:)<\'@(()]\\"<;"(>[());[@@:][:;,]`',
        ';``@[,,[@<;;,]<[";\'`["":;<[(;:])[;@\'\'>@<():`)""<"<\\,@><@@)\']<@"@`]\'(@>(<@@@\'`\'":',
        '@\'[<,[>:@@[`(;:[<<](:@)<<,>">"[\\,:\'@\']\';,\\\\"(]:<,`<",>],\'>`:[,)(@([:>\\\\@]):@\\;,>',
        '))"]["(@,[[\']]<);)\'<)`@,@:"`\'`">"<[\'])\';;("]:,"\\(]@>@\'`;:()(`>)>[]>)<<,>:;,:;@[;',
        ':)]](,);\\\'(\\,,:\\`(<`<\'[@(,;[;<""""`<<(,@@",)`><@[)\'`"`]<]<\\\\:(`\\`>`<;`;(["\\,[[[]',
    ])
def test_sub_masked_for_bwidm_eppn_rnd(sub):
    info = UserInfo({'user': {'userinfo': {'sub': sub}}})
    assert info._sub_masked_for_bwidm_eppn() == "".join(repeat('-', 80))


@pytest.mark.parametrize('iss', [
        "example.org",
        "http://example.org",
        "https://example.org"
    ])
def test_iss_masked_for_bwidm_eppn_fixes_prefix(iss):
    info = UserInfo({'user': {'userinfo': {'iss': iss}}})
    assert info._iss_masked_for_bwidm_eppn() == "example.org"


@pytest.mark.parametrize('raw,cooked', [
        ("exämple.org","example.org"),
        ("example.örg","example.org"),
        ("ürsula.org","ursula.org"),
    ])
def test_iss_masked_for_bwidm_eppn_fixes_umlaute(raw,cooked):
    info = UserInfo({'user': {'userinfo': {'iss': raw}}})
    assert info._iss_masked_for_bwidm_eppn() == cooked


@pytest.mark.parametrize('raw,cooked', [
        ("example.org/foobar","example.org-foobar"),
        ("example.org/foo%20bar","example.org-foo-20bar"),
    ])
def test_iss_masked_for_bwidm_eppn_fixes_urls(raw,cooked):
    info = UserInfo({'user': {'userinfo': {'iss': raw}}})
    assert info._iss_masked_for_bwidm_eppn() == cooked


@pytest.mark.parametrize('raw,cooked', [
        ("fooBarBaz", "foo_bar_baz"),
        ("FooBarBaz", "foo_bar_baz"),
    ])
def test_group_masked_for_bwidm_converts_camel_to_snake_case(raw,cooked):
    info = UserInfo({'user': {'userinfo': {}}})
    assert info._group_masked_for_bwidm(raw) == cooked


@pytest.mark.parametrize('raw,cooked', [
        ("FOOBARBAZ", "foobarbaz"),
        ("FOO-BAR-BAZ", "foo-bar-baz"),
    ])
def test_group_masked_for_bwidm_all_caps(raw,cooked):
    info = UserInfo({'user': {'userinfo': {}}})
    assert info._group_masked_for_bwidm(raw) == cooked


@pytest.mark.parametrize('raw,cooked', [
        ("42", "four_2"),
        ("--test--", "test--"),
        ("__init__()", "init__--"),
        ("?!#_bullshit", "bullshit")
    ])
def test_group_masked_for_bwidm_fixes_beginning(raw,cooked):
    info = UserInfo({'user': {'userinfo': {}}})
    assert info._group_masked_for_bwidm(raw) == cooked


@pytest.mark.parametrize('data', settings.ALL_INPUT)
def test_given_name(userinfo):
    assert userinfo.given_name == "Marcus"


@pytest.mark.parametrize('data', settings.ALL_INPUT)
def test_family_name(userinfo):
    assert userinfo.family_name == "Hardt"


@pytest.mark.parametrize('data', settings.ALL_INPUT)
def test_full_name(userinfo):
    assert userinfo.full_name == "Marcus Hardt"


@pytest.mark.parametrize('data,username', [
        (settings.INPUT_UNITY, "marcus"),
        (settings.INPUT_EGI, "mhardt"),
        (settings.INPUT_DEEP_IAM, "marcus"),
        (settings.INPUT_INDIGO_IAM, "marcus"),
        (settings.INPUT_KIT, "lo0018")
    ])
def test_username(userinfo, username):
    assert userinfo.username == username


@pytest.mark.parametrize('data,email', [
        (settings.INPUT_UNITY, "marcus.hardt@kit.edu"),
        (settings.INPUT_EGI, "marcus.hardt@kit.edu"),
        (settings.INPUT_DEEP_IAM, None),
        (settings.INPUT_INDIGO_IAM, None),
        (settings.INPUT_KIT, "marcus.hardt@kit.edu")
    ])
def test_email(userinfo, email):
    assert userinfo.email == email


@pytest.mark.parametrize('data,unique_id', [
        (settings.INPUT_UNITY, "6c611e2a-2c1c-487f-9948-c058a36c8f0e@https%3A%2F%2Flogin.helmholtz-data-federation.de%2Foauth2"),
        (settings.INPUT_EGI, "d7a53cbe3e966c53ac64fde7355956560282158ecac8f3d2c770b474862f4756%40egi.eu@https%3A%2F%2Faai.egi.eu%2Foidc%2F"),
        (settings.INPUT_DEEP_IAM, "d9730f60-3b19-4f45-83ab-f29addf72d58@https%3A%2F%2Fiam.deep-hybrid-datacloud.eu%2F"),
        (settings.INPUT_INDIGO_IAM, "a1ea3aa2-8daf-41bb-b4fb-eb88f439e446@https%3A%2F%2Fiam-test.indigo-datacloud.eu%2F"),
        (settings.INPUT_KIT, "4cbcd471-1f51-4e54-97b8-2dd5177e25ec@https%3A%2F%2Foidc.scc.kit.edu%2Fauth%2Frealms%2Fkit%2F")
    ])
def test_unique_id(userinfo, unique_id):
    assert userinfo.unique_id == unique_id


@pytest.mark.parametrize('data,eppn', [
        (settings.INPUT_UNITY, "6c611e2a-2c1c-487f-9948-c058a36c8f0e@login.helmholtz-data-federation.de-oauth2"),
        (settings.INPUT_EGI, "d7a53cbe3e966c53ac64fde7355956560282158ecac8f3d2c770b474862f4756-egi.eu@aai.egi.eu-oidc-"),
        (settings.INPUT_DEEP_IAM, "d9730f60-3b19-4f45-83ab-f29addf72d58@iam.deep-hybrid-datacloud.eu-"),
        (settings.INPUT_INDIGO_IAM, "a1ea3aa2-8daf-41bb-b4fb-eb88f439e446@iam-test.indigo-datacloud.eu-"),
        (settings.INPUT_KIT, "4cbcd471-1f51-4e54-97b8-2dd5177e25ec@oidc.scc.kit.edu-auth-realms-kit-")
    ])
def test_eppn(userinfo, eppn):
    assert userinfo.eppn == eppn


@pytest.mark.parametrize('data,groups', [
        (settings.INPUT_UNITY, ["h-df-de_imk-tro-ewcc", "h-df-de_my_example_colab", "h-df-de_wlcg-test", "h-df-de_hdf"]),
        (settings.INPUT_EGI, []),
        (settings.INPUT_DEEP_IAM, ["kit-cloud"]),
        (settings.INPUT_INDIGO_IAM, ["users", "developers", "test-vo-users"]),
        # (settings.INPUT_KIT, ["kit-edu_dfn-slcs", "kit-edu_lsdf-dis", "kit-edu_bwgrid", "kit-edu_bwlsdf-fs",
        #                       "kit-edu_bwunicluster", "kit-edu_bwsyncandshare", "kit-edu_bwsyncandshare-idm", "kit-edu_grruppenverwalter"])
    ])
def test_groups(userinfo, groups):
    # TODO: check why KIT test fails (seems the entitlements in the input are invalid?)
    assert sorted(userinfo.groups) == sorted(groups)


@pytest.mark.parametrize('data,group', [
    (settings.INPUT_UNITY, "h-df-de_hdf"),
    (settings.INPUT_EGI, None),
    (settings.INPUT_DEEP_IAM, "kit-cloud"),
    (settings.INPUT_INDIGO_IAM, "developers"),
    (settings.INPUT_KIT, None),
])
def test_primary_group_no_fallback_or_primary_configured(userinfo, group):
    assert userinfo.primary_group == group


@pytest.mark.parametrize('data,group', [
    (settings.INPUT_UNITY, "h-df-de_hdf"),
    (settings.INPUT_EGI, "nogroup"),
    (settings.INPUT_DEEP_IAM, "kit-cloud"),
    (settings.INPUT_INDIGO_IAM, "developers"),
    (settings.INPUT_KIT, "nogroup"),
])
def test_primary_group_fallback_configured_no_primary(userinfo, group, monkeypatch):
    monkeypatch.setitem(CONFIG['ldf_adapter'], "fallback_group", "nogroup")
    assert userinfo.primary_group == group


@pytest.mark.parametrize('data', settings.ALL_INPUT)
def test_primary_group_primary_and_fallback_configured(userinfo, monkeypatch):
    monkeypatch.setitem(CONFIG['ldf_adapter'], "fallback_group", "nogroup")
    monkeypatch.setitem(CONFIG['ldf_adapter'], "primary_group", "mytestgroup")
    assert userinfo.primary_group == "mytestgroup"


def test_egi_sub_is_unscoped():
    assert '@' in settings.INPUT_EGI['user']['userinfo']['sub']


@pytest.mark.parametrize('data,assurance', [
    (settings.INPUT_UNITY, ["https://refeds.org/assurance/IAP/medium",
            "https://refeds.org/assurance/IAP/local-enterprise",
            "https://refeds.org/assurance/ID/eppn-unique-no-reassign",
            "https://refeds.org/assurance/ATP/ePA-1m",
            "https://refeds.org/assurance/ATP/ePA-1d",
            "https://refeds.org/assurance/ID/unique",
            "https://refeds.org/assurance/profile/cappuccino",
            "https://refeds.org/assurance/IAP/low"]),
    (settings.INPUT_EGI, ["https://aai.egi.eu/LoA#Substantial"]),
    (settings.INPUT_DEEP_IAM, []),
    (settings.INPUT_INDIGO_IAM, []),
    (settings.INPUT_KIT, [])
])
def test_missing_assurance(userinfo, assurance):
    assert sorted(userinfo.assurance) == sorted(assurance)


def test_ignore_excess_entitlement():
    """https://git.scc.kit.edu/feudal/feudalAdapterLdf/issues/8"""

    input_test = {
        "eduperson_assurance": [
            "https://refeds.org/assurance/IAP/medium",
            "https://refeds.org/assurance/IAP/local-enterprise",
            "https://refeds.org/assurance/ID/eppn-unique-no-reassign",
            "https://refeds.org/assurance/ATP/ePA-1m",
            "https://refeds.org/assurance/ATP/ePA-1d",
            "https://refeds.org/assurance/ID/unique",
            "https://refeds.org/assurance/profile/cappuccino",
            "https://refeds.org/assurance/IAP/low"
        ],
        "eduperson_entitlement": [
            "urn:mace:dir:entitlement:common-lib-terms",
            "http://bwidm.de/entitlement/bwLSDF-SyncShare",
            "urn:geant:h-df.de:group:IMK-TRO-EWCC#login.helmholtz-data-federation.de",
            "urn:geant:h-df.de:group:MyExampleColab#login.helmholtz-data-federation.de",
            "urn:geant:h-df.de:group:wlcg-test#login.helmholtz-data-federation.de",
            "urn:geant:h-df.de:group:HDF#login.helmholtz-data-federation.de"
        ],
        "eduperson_scoped_affiliation": "member@kit.edu",
        "email": "marcus.hardt@kit.edu",
        "email_verified": "true",
        "family_name": "Hardt",
        "given_name": "Marcus",
        "groups": [
            "/wlcg-test",
            "/IMK-TRO-EWCC",
            "/MyExampleColab",
            "/HDF",
            "/"
        ],
        "iss": "https://login.helmholtz-data-federation.de/oauth2",
        "name": "Marcus Hardt",
        "preferred_username": "marcus",
        "ssh_key": "ssh-ed25519 AAAAC3NzaC1lZDI1NTE5AAAAIAqA5FW6m3FbFhCOsRQBxKMRki5qJxoNhZdaeLXg6ym/ marcus@nemo2019\n",
        "sub": "6c611e2a-2c1c-487f-9948-c058a36c8f0e"
    }

    info = UserInfo({'user': {'userinfo': input_test}})
    assert len(list(info.entitlement)) == 4




import pytest
import subprocess
import os
from pathlib import Path

from ldf_adapter import UserInfo, User, backend, CONFIG
import logging

logger = logging.getLogger(__name__)

@pytest.fixture(scope="function")
def userinfo(data):
    """Creates a UserInfo object from provided dict"""
    userinfo = UserInfo(data)
    yield userinfo


@pytest.fixture(scope="function")
def user(data, monkeypatch):
    """Creates a User from provided dict"""
    # monkeypatch.setitem(CONFIG['ldf_adapter'], "backend", "local_unix")
    user = User(data)
    yield user


@pytest.fixture(scope="function")
def local_unix_user(input, exists, taken, monkeypatch):
    """Creates a backend user from provided dict data.
    input should contain:
        - userinfo, which should contain: unique_id, username, primary_group and ssh_keys.
        - new_root: the folder relative to which the user and group dbs will be stored
    If exists, it also adds an entry to the user database in /etc/passwd
    Otherwise, if taken, it adds an entry in the user db for this user's username
    """
    # save original subprocess.run for calling inside the mocked one
    old_subprocess_run = subprocess.run

    def mock_root():
        return input["new_root"]

    def mock_subprocess_run(*args, **kwargs):
        """patches calls to system utilities:
        - only for: useradd, userdel, usermod, chage
        - add prefix argument to command
        - patch pkill to do nothing
        - lett all other system calls go through
        """
        logger.debug(args)
        command = args[0]
        if command[0] in ["useradd", "userdel", "usermod", "chage"]:
            new_command = [command[0], "--prefix", mock_root()] + command[1:]
        elif command[0] == "/usr/bin/pkill":
            return None
        else:
            new_command = command
        return old_subprocess_run(new_command, *args[1:], **kwargs)

    class MockUserInfo():
        """Mocks a UserInfo object to be passed to local_unix.User
        Only a few properties are necessary:
            - unique_id
            - username
            - primary_group
            - ssh_keys
        """
        def __init__(self, data):
            self.unique_id = data["unique_id"]
            self.username = data["username"]
            self.primary_group = data["primary_group"]
            self.ssh_keys = data["ssh_keys"]
    
    monkeypatch.setitem(CONFIG['ldf_adapter'], "backend", "local_unix")
    monkeypatch.setitem(CONFIG['backend.local_unix'], "shell", "/bin/bash")
    monkeypatch.setattr("subprocess.run", mock_subprocess_run)
    backend.User.ROOT = mock_root  # type: ignore
    backend.Group.ROOT = mock_root  # type: ignore

    # init root and necessary files in new root directory (/etc/{passwd,group,shadow})
    os.makedirs(mock_root())
    os.makedirs(Path(mock_root())/"etc")
    (Path(mock_root())/"etc"/"passwd").touch()
    (Path(mock_root())/"etc"/"group").touch()
    (Path(mock_root())/"etc"/"shadow").touch()

    if exists:
        (Path(mock_root())/"etc"/"passwd").write_text(input["passwd_entry"])
    elif taken:
        (Path(mock_root())/"etc"/"passwd").write_text(input["passwd_taken"])
    (Path(mock_root())/"etc"/"group").write_text(input["group_entry"])


    # init service user from unix backend
    service_user = backend.User(MockUserInfo(input["userinfo"]))  # type:ignore

    yield service_user

    # clean up files
    old_subprocess_run(['rm', '-rf', mock_root()])


@pytest.fixture(scope="function")
def local_unix_group(input, exists, monkeypatch):
    """Creates a backend user from provided dict data.
    input should contain:
        - name: the group name (no constraints on allowed names)
        - new_root: the folder relative to which the user and group dbs will be stored
    If exists=True, also adds an entry to the group database in /etc/group
    """
    # save original subprocess.run for calling inside the mocked one
    old_subprocess_run = subprocess.run

    def mock_root():
        return input["new_root"]

    def mock_subprocess_run(*args, **kwargs):
        """patches calls to system utilities:
        - only for: groupadd
        - add prefix argument to command
        - lett all other system calls go through
        """
        logger.debug(args)
        command = args[0]
        if command[0] in ["groupadd"]:
            new_command = [command[0], "--prefix", mock_root()] + command[1:]
        else:
            new_command = command
        return old_subprocess_run(new_command, *args[1:], **kwargs)

    monkeypatch.setitem(CONFIG['ldf_adapter'], "backend", "local_unix")
    monkeypatch.setattr("subprocess.run", mock_subprocess_run)
    backend.Group.ROOT = mock_root  # type:ignore

    # init root and necessary files in new root directory (/etc/{passwd,group,shadow})
    os.makedirs(mock_root())
    os.makedirs(Path(mock_root())/"etc")
    (Path(mock_root())/"etc"/"group").touch()

    if exists:
        (Path(mock_root())/"etc"/"group").write_text(input["group_entry"])

    # init service user from unix backend
    service_group = backend.Group(input["name"])  # type:ignore

    yield service_group

    # clean up files
    old_subprocess_run(['rm', '-rf', mock_root()])



class MockBackendUserDB():
    """Simple user db represented as a dict"""
    def __init__(self):
        pass


class MockBackendUser():
    """Mock user for the backend"""
    def __init__(self):
        pass

# vim: tw=100



name = "ldf_adapter"
# vim: tw=100 foldmethod=expr
# pylint: disable=invalid-name, superfluous-parens
# pylint: disable=redefined-outer-name, logging-not-lazy, logging-format-interpolation, logging-fstring-interpolation
# pylint: disable=missing-docstring, trailing-whitespace, trailing-newlines, too-few-public-methods

import logging
import sys
from feudal_globalconfig import globalconfig

import regex

# from unidecode import unidecode

from . import logsetup

# from . import eduperson

from . import backend
from .config import CONFIG
from .results import (
    Deployed,
    NotDeployed,
    Rejection,
    Failure,
    Question,
    raise_question,
    Status,
)
from .name_generators import FriendlyNameGenerator, PooledNameGenerator, NameGenerator
from .userinfo import UserInfo

logger = logging.getLogger(__name__)


class User:
    """Represents a user, abstracting from the concrete service.

    An abstract User is backed by a service_user and is associated with a set of groups (backed by
    service_groups).

    A user is usually identified on the service not by a username but by `self.data.unique_id` (see
    __init__ for details).
    """

    def __init__(self, data):
        """
        Arguments:
        data -- Information about the user (type: UserInfo or dict)

        Relevant config:
        ldf_adapter.backend -- The name of the backend. See function the `backend` for possible values
        ldf_adapter.primary_group -- The primary group of the user. If empty, one from the
          supplementary groups will be used. If there are multiple, a question will be raised.

        Words of warning: Since the service_user and service_groups are
        backend specific, their structures differ from backend to backend.

        Direct access to them from this __init__ is highly illegal (unless specified)
        Instead: Use self.data
        """
        # Info Display Hack
        if CONFIG.getboolean("verbose-info-plugin", "active", fallback=False) is True:
            try:
                if data["state_target"] == "deployed":
                    import json
                    import os
                    import stat

                    filename = CONFIG.get(
                        "verbose-info-plugin",
                        "filename",
                        fallback="/tmp/userinfo/userinfo.json",
                    )
                    dirname = CONFIG.get(
                        "verbose-info-plugin", "dirname", fallback="/tmp/userinfo"
                    )
                    try:
                        os.mkdir(dirname)
                        os.chmod(dirname, 0o0777)
                    except:
                        pass
                    with open(filename, "w", encoding="utf-8") as f:
                        json.dump(
                            data,
                            f,
                            ensure_ascii=False,
                            sort_keys=True,
                            indent=4,
                            separators=(",", ": "),
                        )
                    os.chmod(filename, 0o0666)
                    # os.chmod(filename, stat.S_IREAD || stat.S_IWRITE ||stat.S_IRGRP || stat.S_IWGRP
                    #         || stat.S_IROTH || stat.S_IWOTH)
            except Exception as e:
                logger.error(f"Got an exception in uncritical code: {e}")

        # Proceed as normal
        self.data = data if isinstance(data, UserInfo) else UserInfo(data)
        self.service_user = backend.User(self.data)  # type: ignore
        self.service_groups = [
            backend.Group(grp) for grp in self.data.groups  # type:ignore
        ]

        if CONFIG.get(
            "ldf_adapter", "backend_supports_preferring_existing_user", fallback=False
        ):
            logger.debug("trying to update user from existing")
            if self.service_user.exists():
                self.update_username_from_existing()

    def assurance_verifier(self):
        """Produce a suitably function to check if a user is allowed.

        Relevant config:
        assurance.prefix -- The common prefix of all relative assurance claims
        assurance.require -- The boolean expression to be parsed, according to the following
          grammer: `E -> E "&" E | E "|" E | "(" E ")" | string`, where `&` binds stronger than `|`.
          Strings are assurance claims interpreted absolute (if they start with `"http[s]://"`) or
          relative to `assurance.prefix`. The strings "+" and "*" are interpreted specially: "+"
          means "any assurance claim", while "*" means "any claim, or no claim at all". They thus
          differ in their treatment of users without any claims.

        Returns:
        A function taking a set of assurance claims, interpreted absolute. The function returns
        `True`, if the claims satisfy the configured expression (`"assurance.require"`), `False`
        otherwise.
        """
        ass = CONFIG["assurance"]
        prefix = ass["prefix"]
        prefix = prefix.rstrip("/") + "/"

        tokens = regex.findall("&|\||\(|\)|[^\s()&|]+", ass["require"])

        # We use a simple recursive descent parser to parse parenthesied expressions of strings,
        # composed with '&' (konjunction) and '|' (disjunction). The usual precedence rules apply.
        #
        # Instead of building an AST, we build a tree of nested lambdas, which takes a collection of
        # assurance claims and checks if they satisfy the configured expression

        #  EXPR -> DISJ 'EOF'
        def parse_expr(seq):
            expr = parse_disjunction(seq)
            if len(seq) > 0:
                raise ValueError("Reached end of input while parsing")
            return expr

        #  DISJ -> KONJ DISJ2
        def parse_disjunction(seq):
            lhs = parse_konjunction(seq)
            return parse_disjunction2(seq, lhs)

        #  DISJ2 -> ""
        #        -> "|" KONJ DISJ2
        def parse_disjunction2(seq, lhs):
            if len(seq) > 0 and seq[0] == "|":
                seq.pop(0)
                rhs = parse_konjunction(seq)
                expr = lambda values: lhs(values) or rhs(values)
                return parse_disjunction2(seq, expr)
            else:
                return lhs

        #  KONJ -> PRIMARY KONJ2
        def parse_konjunction(seq):
            lhs = parse_primary(seq)
            return parse_konjunction2(seq, lhs)

        #  KONJ2 -> ""
        #        -> "&" PRIMARY
        def parse_konjunction2(seq, lhs):
            if len(seq) > 0 and seq[0] == "&":
                seq.pop(0)
                rhs = parse_primary(seq)
                expr = lambda values: lhs(values) and rhs(values)
                return parse_konjunction2(seq, expr)
            else:
                return lhs

        #  PRIMARY -> "(" DISJ ")"
        #          -> ASSURANCE
        def parse_primary(seq):
            if len(seq) > 0 and seq[0] == "(":
                seq.pop(0)
                subexpr = parse_disjunction(seq)
                if len(seq) > 0 and seq.pop(0) != ")":
                    raise ValueError("Missing ')' while parsing")
                return subexpr
            else:
                return parse_assurance(seq)

        #  ASSURANCE -> string
        #            -> "*"
        #            -> "+"
        def parse_assurance(seq):
            value = seq.pop(0)
            if value == "+":
                return lambda values: len(values) > 0
            elif value == "*":
                return lambda values: True
            else:
                value = value if regex.match("https?://", value) else prefix + value
                return lambda values: value in values

        return parse_expr(tokens)

    def reach_state(self, target):
        """Attempt to put the user into the desired state on the configured service.

        Arguments:
        target -- The desired state. One of 'deployed' and 'not_deployed'.
        user -- The user to be deployed/undeployed (type: User)
        """

        username = "not yet assigned"
        if self.service_user.exists():
            username = self.service_user.get_username()
        try:
            logger.info(
                f"Incoming request to reach '{target}' for user with email: '{self.data.email}' ({self.data.unique_id}) username: {username}"
            )
        except AttributeError:
            logger.info(
                f"Incoming request to reach '{target}' for user with name: '{self.data.full_name}' ({self.data.unique_id}) username: {username}"
            )

        if target == "deployed":
            if CONFIG.get("assurance", "skip", fallback="No") == "Yes, do as I say!":
                logger.warning(
                    "Assurance checking is disabled: Users with ANY assurance will be authorised"
                )
            if (
                not CONFIG.get("assurance", "skip", fallback="No")
                == "Yes, do as I say!"
            ):
                if not self.assurance_verifier()(self.data.assurance):
                    raise Rejection(
                        message="Your assurance level is insufficient to access this resource"
                    )

            logger.debug(f"User comes with these groups")
            for g in self.service_groups:
                logger.debug(f"    {g.name}")

            return self.deploy()
        elif target == "not_deployed":
            if (
                not CONFIG.get("assurance", "skip", fallback="No")
                == "Yes, do as I say!"
            ):
                if not self.assurance_verifier()(self.data.assurance):
                    if not CONFIG.getboolean(
                        "assurance", "verified_undeploy", fallback=False
                    ):
                        logger.warning(
                            "Assurance level is insufficient. Undeploying anyway."
                        )
                    else:
                        raise Rejection(
                            message="Your assurance level is insufficient to access this resource"
                        )
            return self.undeploy()
        elif target == "get_status":
            return self.get_status()
        elif target == "resumed":
            return self.resume()
        elif target == "suspended":
            return self.suspend()
        elif target == "limited":
            return self.limit()
        else:
            raise ValueError(f"Invalid target state: {target}")

    def deploy(self):
        """Deploy the user.

        Ensure that the user exists, is a member in the right groups (and only in those groups)
        and has the correct credentials installed.

        Return a Deployed result, with a message describing what was done.
        """
        self.ensure_groups_exist()
        was_created = self.ensure_exists()
        new_groups = self.ensure_group_memberships()
        new_credentials = self.ensure_credentials_active()

        what_changed = ""
        if was_created:
            what_changed += "User was created"
        else:
            # FIXME: a user that was not created might not exist for other reasons.
            #        the code probably relies on Failures rosen.
            #        A "pending" flow might require additional Classes for return
            what_changed += "User already existed"

        if new_groups:
            what_changed += " and was added to groups {}".format(",".join(new_groups))

        what_changed += "."

        if new_credentials:
            what_changed += " Credentials {} were activated.".format(
                ",".join(new_credentials)
            )

        return Deployed(credentials=self.credentials, message=what_changed)

    def undeploy(self):
        """Ensure that the user dosen't exist.

        Return a NotDeployed result with a message saying if the user previously existed.
        """
        username = self.service_user.get_username()
        was_removed = self.ensure_dosent_exist()

        what_changed = ""
        if was_removed:
            what_changed += f"User '{username} ({self.data.unique_id})' was removed."
        else:
            what_changed += (
                f"No user for '{self.data.unique_id}' existed. "
                + f"User '{username}' was not changed"
            )

        return NotDeployed(message=what_changed)

    def suspend(self):
        """Ensure that the user is suspended.

        Return a Status result with a message describing what was done.
        """
        was_suspended = self.ensure_suspended()
        what_changed = ""
        if was_suspended:
            # FIXME: I'm not sure if this ought to be self.data.username
            what_changed += f"User '{self.data.unique_id}' was suspended."
            state = "suspended"
        else:
            state = self.get_status().state
            what_changed += (
                f"Suspending user '{self.data.unique_id}' was not possible from the '{state}' state. "
                + f"User was not changed."
            )
        return Status(state, message=what_changed)

    def resume(self):
        """Ensure that a suspended user is active again in state 'deployed'.

        Return a Status result with a message describing what was done.
        """
        was_resumed = self.ensure_resumed()
        what_changed = ""
        state = self.get_status().state
        if was_resumed:
            # FIXME: I'm not sure if this ought to be self.data.username
            what_changed += f"User '{self.data.unique_id}' was resumed."
        else:
            what_changed += (
                f"Resuming user '{self.data.unique_id}' was not possible from the '{state}' state. "
                + f"User was not changed."
            )
        return Status(state, message=what_changed)

    def limit(self):
        """Ensure that a user has limited access.

        Return a Status result with a message describing what was done.
        """
        was_limited = self.ensure_limited()
        what_changed = ""
        if was_limited:
            # FIXME: I'm not sure if this ought to be self.data.username
            what_changed += f"User '{self.data.unique_id}' was limited."
            state = "limited"
        else:
            state = self.get_status().state
            what_changed += (
                f"Limiting user '{self.data.unique_id}' was not possible from the '{state}' state. "
                + f"User was not changed."
            )
        return Status(state, message=what_changed)

    def unlimit(self):
        """Ensure that a limited user is active again in state 'deployed'.

        Return a Status result with a message describing what was done.
        """
        was_unlimited = self.ensure_unlimited()
        what_changed = ""
        if was_unlimited:
            # FIXME: I'm not sure if this ought to be self.data.username
            what_changed += f"User '{self.data.unique_id}' was unlimited."
            state = "deployed"
        else:
            state = self.get_status().state
            what_changed += (
                f"Resuming user '{self.data.unique_id}' was not possible from the '{state}' state. "
                + f"User was not changed."
            )
        return Status(state, message=what_changed)

    def get_status(self):
        """
        Return the current status (that he has in the underlying local user management system)
        User can have these status:
        +--------------+-----------------------------------------------------------------+-----------------+
        | Status       | Comment                                                         | Backend support |
        +--------------+-----------------------------------------------------------------+-----------------+
        +--------------+-----------------------------------------------------------------+-----------------+
        | deployed     | There is an account for the user identified by unique_id        | Mandatory       |
        +--------------+-----------------------------------------------------------------+-----------------+
        | not deployed | There is no account for the user identified by unique_id        | Mandatory       |
        |              | We have no information if there has ever been an account        |                 |
        +--------------+-----------------------------------------------------------------+-----------------+
        | rejected     | This might not be supportable; Depends on the backend           | Optional        |
        +--------------+-----------------------------------------------------------------+-----------------+
        | suspended    | The user with unique_id has been suspended                      | Optional        |
        +--------------+-----------------------------------------------------------------+-----------------+
        | pending      | The creation of the user is pending                             | Optional        |
        +--------------+-----------------------------------------------------------------+-----------------+
        | limited      | The user was limited, typically after being idle for some time  | Optional        |
        +--------------+-----------------------------------------------------------------+-----------------+
        | unknown      | We don't know the status, but at least the user is not deployed | Mandatory       |
        +--------------+-----------------------------------------------------------------+-----------------+
        """

        msg = "No message"
        try:
            if not self.service_user.exists():
                return Status("not_deployed", message=msg)
            msg = f"username {self.service_user.get_username()}"
            if hasattr(self.service_user, "is_rejected"):
                if self.service_user.is_rejected():
                    return Status("rejected", message=msg)
            if hasattr(self.service_user, "is_suspended"):
                if self.service_user.is_suspended():
                    return Status("suspended", message=msg)
            if hasattr(self.service_user, "is_pending"):
                if self.service_user.is_pending():
                    return Status("pending", message=msg)
            if hasattr(self.service_user, "is_limited"):
                if self.service_user.is_limited():
                    return Status("limited", message=msg)
            return Status("deployed", message=msg)
        except Exception as e:
            logger.error(f"User {self.data.unique_id} is in an undefined state.: {e}")
            return Status("unknown", message=msg)

    def ensure_exists(self):
        """Ensure that the user exists on the service.

        If the username is already taken on the service, raise a questionaire for a new one. See
        UserInfo.username for details.

        Also ensure that all info about the user is up to date on the service. This is done
        independently of creating the user, so that the user is updated even if they already existed.

        Return True, if the user didn't exist before.
        """
        logger.debug(f"Ensuring a local user mapping for {self.data.unique_id} exits")

        is_new_user = not self.service_user.exists()

        unique_id = self.data.unique_id
        if is_new_user:
            username = self.data.username
            primary_group_name = self.data.primary_group

            # Raise question in case of existing username in case we're interactive
            if CONFIG.getboolean(
                "ldf_adapter", "interactive", fallback=False
            ):  # interactive
                logger.debug("interactive mode")
                if self.service_user.name_taken(username):
                    logger.info(
                        f'Username "{username}" is already taken, asking user to pick a new one'
                    )
                    raise Question(
                        name="username",
                        text=f'Username "{username}" already taken on this service. Please enter another one.',
                    )

            else:  # non-interactive
                logger.debug("noninteractive mode")
                username_mode = CONFIG.get(
                    "username_generator", "mode", fallback="friendly"
                )
                logger.debug(f"username_mode: {username_mode}")
                pool_prefix = CONFIG.get(
                    "username_generator", "pool_prefix", fallback=primary_group_name
                )

                name_generator = NameGenerator(
                    username_mode, userinfo=self.data, pool_prefix=pool_prefix
                )
                proposed_name = name_generator.suggest_name()
                logger.debug(f"initially proposed_name: {proposed_name}")

                while self.service_user.name_taken(proposed_name):
                    proposed_name = name_generator.suggest_name()
                if proposed_name is None:
                    raise Rejection(
                        message=f"I cannot create usernames. "
                        f"The list of tried ones is: {', '.join(name_generator.tried_names())}."
                    )
                self.service_user.set_username(proposed_name)

                logger.info(f"Chose username '{proposed_name}' for {unique_id}")

            logger.debug(f"Primary Group Name: {primary_group_name}")
            logger.debug(f"Primary Group from userinfo: {self.data.primary_group}")

            # Sanity check to ensure user has a primary group:
            if primary_group_name is None:
                config_file_name = globalconfig.info["config_files_read"]
                message = (
                    'User is not member of any group, and neither a "primary_group", nor '
                    f'a "fallback_group" have been defined in the config file {config_file_name}'
                )
                logger.error(message)
                print(f"\nERROR: {message}\n")
                sys.exit(2)

            self.service_user.create()
        else:  # The user exists
            # Update service_user.name if unique_id already points to a username:
            username = self.service_user.get_username()
            logger.info(f"User {username} for '{self.data.unique_id}' already exists.")

        logger.debug(f"This is a new user: {is_new_user}")

        self.service_user.update()
        return is_new_user

    def update_username_from_existing(self):
        """Update self.service_user.name, if a user with matching
        unique_id can be found, and if the backend implements 'set_username'
        """
        try:
            existing_username = self.service_user.get_username()
            if existing_username is not None:
                if hasattr(self.service_user, "set_username"):
                    logger.debug(
                        f"Setting username to {existing_username} ({self.data.unique_id})"
                    )
                    self.service_user.set_username(existing_username)
                logger.debug(f"Found an existing username: {existing_username}")
        except AttributeError:
            # the currently used service_user class has to method get_username
            existing_username = None

    def ensure_dosent_exist(self):
        """Ensure that the user doesn't exist.

        Before deleting them, uninstall all SSH keys, to be sure that they are really gone.

        Return True, if the user existed before.
        """
        if self.service_user.exists():
            self.service_user.username = self.service_user.get_username()
            logger.info(
                f"Deleting user '{self.service_user.username}' ({self.data.unique_id})"
            )
            # bwIDM requires prior removal of the user, because ssh-key removal triggers an
            # asyncronous process. If user is removed during that, the user might be only partially
            # removed...
            if CONFIG.get("ldf_adapter", "backend", fallback="") == "bwidm":
                self.service_user.delete()
                self.service_user.uninstall_ssh_keys()
            else:
                self.service_user.uninstall_ssh_keys()
                self.service_user.delete()
            return True
        else:
            logger.info(
                f"No user existed for {self.data.unique_id} did exist. Nothing to do."
            )
            return False

    def ensure_suspended(self):
        """Ensure that a user is suspended.
        Return True if the user has been suspended.
        """
        status = self.get_status()
        if status.state in ["deployed", "limited"]:
            if hasattr(self.service_user, "suspend"):
                self.service_user.suspend()
                return True
        logger.debug(
            f"User {self.data.unique_id} in state {status.state}. Suspending not allowed."
        )
        return False

    def ensure_limited(self):
        """Ensure that a user has limited access.
        Return True if setting the user is limited.
        """
        status = self.get_status()
        if status.state == "deployed":
            if hasattr(self.service_user, "limit"):
                self.service_user.limit()
                return True
        logger.debug(
            f"User {self.data.unique_id} in state {status.state}. Limiting not allowed."
        )
        return False

    def ensure_resumed(self):
        """Ensure that a user is resumed.
        Return True is the user
        """
        status = self.get_status()
        if status.state == "suspended":
            if hasattr(self.service_user, "resume"):
                self.service_user.resume()
                return True
        logger.debug(
            f"User {self.data.unique_id} in state {status.state}. Resuming not allowed."
        )
        return False

    def ensure_unlimited(self):
        """Ensure that a user is not limited anymore.
        Return True is the user
        """
        status = self.get_status()
        if status.state == "limited":
            if hasattr(self.service_user, "unlimit"):
                self.service_user.unlimit()
                return True
        logger.debug(
            f"User {self.data.unique_id} in state {status.state}. Unlimit not allowed."
        )
        return False

    def ensure_groups_exist(self):
        """Ensure that all the necessary groups exist.

        Create the groups on the service, if necessary.
        """
        group_list = filter(
            lambda grp: not grp.exists(),
            [self.service_user.primary_group] + self.service_groups,
        )
        for group in group_list:
            if group.name is not None:
                logger.info("Creating group '{}'".format(group.name))
                group.create()

    def ensure_group_memberships(self):
        """Ensure that the user is a member of all the groups in self.service_groups.

        Return the names of all groups the user is now a member of.
        """
        # Note: current code will keep on adding the user to the primary group. This may be a
        # problem with some backends

        group_list = self.service_groups
        if self.service_user.primary_group.name not in [
            grp.name for grp in self.service_groups
        ]:
            group_list.append(self.service_user.primary_group)

        if group_list[0].name is None:
            config_file_name = globalconfig.info["config_files_read"]
            message = (
                'User is not member of any group, and neither a "primary_group", nor '
                f'a "fallback_group" have been defined in the config file {config_file_name}'
            )
            logger.error(message)
            print(f"\nERROR: {message}\n")
            sys.exit(3)
        username = self.service_user.get_username()
        logger.info(
            f"Ensuring user '{username}' ({self.data.unique_id}) is member of these groups: {[grp.name for grp in group_list]}"
        )
        self.service_user.mod(supplementary_groups=group_list)
        return [grp.name for grp in group_list]

    def ensure_credentials_active(self):
        """Install all SSH Keys on the service.

        Return a list of the names/ids of all the keys now active.
        """
        self.service_user.install_ssh_keys()
        return ["ssh:{name}/{id}".format(**key) for key in self.data.ssh_keys]

    @property
    def credentials(self):
        """The Credentials displayed to the user.

        Simply merges all the credentials provided by the service_user with those configured for
        the backend in the config file.

        See Deployed.__init__ for details on how this value is used.

        Relevant config:
        ldf_adapter.backend -- The backend to be used
        backend.{}.login_info -- Everything in this section is merged into the credentials dictionary.
        """
        return {
            **self.service_user.credentials,
            **CONFIG["backend.{}.login_info".format(CONFIG["ldf_adapter"]["backend"])],
        }



"""Simple example backend, for documentation purposes only."""


class ExampleUser:
    """Manages the user object on the service."""

    def __init__(self, userinfo):
        """
        Arguments:
        userinfo -- (type: UserInfo)
        """
        pass

    def exists(self):
        """Return whether the user exists on the service.

        If this returns True,  calling `create` should have no effect or raise an error.
        """
        return bool()

    def name_taken(self, name):
        """Return whether a given username is already taken by another user on the service.

        Should return True if the name is not available for this user (even if it is available
        for other users for some reason)
        """
        return bool()

    def create(self):
        """Create the user on the service.

        If the user already exists, do nothing or raise an error
        """
        pass

    def update(self):
        """Update all relevant information about the user on the service.

        If the user doesn't exists, behaviour is undefined.
        """
        pass

    def delete(self):
        """Delete the user on the service.

        If the user doesn't exists, do nothing or raise an error.
        """
        pass

    def mod(self, supplementary_groups=None):
        """Modify the user on the service.

        The state of the user with respect to the provided Arguments after calling this function
        should not depend on the state the user had previously.

        If the user doesn't exists, behaviour is undefined.

        Arguments:
        supplemantary_groups -- A list of groups to add the user to (type: list(ExampleGroup))
        """
        pass

    def install_ssh_keys(self):
        """Install users SSH keys on the service.

        No other SSH keys should be active after calling this function.

        If the user doesn't exists, behaviour is undefined.
        """
        pass

    def uninstall_ssh_keys(self):
        """Uninstall the users SSH keys on the service.

        This must uninstall all SSH keys installed with `install_ssh_keys`. It may uninstall SSH
        keys installed by other means.

        If the user doesn't exists, behaviour is undefined.
        """
        pass

    def get_username(self):
        """Return local username on the service.

        If the user doesn't exists, behaviour is undefined.
        """
        pass

    def set_username(self, username):
        """Set local username on the service."""
        pass

    @property
    def credentials(self):
        """Return any additional login information required to access the service after deployment.

        This should not include the installed SSH keys.

        If the user is not fully deployed, behaviour is undefined.
        """
        pass

    def is_rejected(self):
        """Optional, only if the backend supports it.
        Return whether a pending user deployment request was rejected"""
        return False

    def is_suspended(self):
        """Optional, only if the backend supports it.
        Return whether the user was suspended (e.g. due to a security incident)"""
        return False

    def is_pending(self):
        """Optional, only if the backend supports it.
        Return whether the user deployment request is pending approval"""
        return False

    def is_limited(self):
        """Optional, only if the backend supports it.
        Return whether the user has limited access"""
        return False

    def suspend(self):
        """Optional, only if the backend supports it.
        Suspends the user such that no access to the service is possible"""
        pass

    def resume(self):
        """Optional, only if the backend supports it.
        Restores the suspended user"""
        pass

    def limit(self):
        """Optional, only if the backend supports it.
        Limits the user's capabilities on the service (e.g. read-only access)"""
        pass

    def unlimit(self):
        """Optional, only if the backend supports it.
        Restores a user with limited access to full capabilities"""
        pass


class ExampleGroup:
    """Manages the group object on the service."""

    def __init__(self, name):
        """
        Arguments:
        name -- The name of the group
        """
        pass

    def exists(self):
        """Return whether the group already exists."""
        pass

    def create(self):
        """Create the group on the service.

        If the group already exists, behaviour is undefined.
        """
        pass




def wrap_userinfo(input):
    return {'user': {'userinfo': input}}


INPUT_UNITY = wrap_userinfo({
        "display_name": "Marcus Hardt",
        "eduperson_assurance": [
            "https://refeds.org/assurance/IAP/medium",
            "https://refeds.org/assurance/IAP/local-enterprise",
            "https://refeds.org/assurance/ID/eppn-unique-no-reassign",
            "https://refeds.org/assurance/ATP/ePA-1m",
            "https://refeds.org/assurance/ATP/ePA-1d",
            "https://refeds.org/assurance/ID/unique",
            "https://refeds.org/assurance/profile/cappuccino",
            "https://refeds.org/assurance/IAP/low"
        ],
        "eduperson_entitlement": [
            "urn:geant:h-df.de:group:IMK-TRO-EWCC#login.helmholtz-data-federation.de",
            "urn:geant:h-df.de:group:MyExampleColab#login.helmholtz-data-federation.de",
            "urn:geant:h-df.de:group:wlcg-test#login.helmholtz-data-federation.de",
            "urn:geant:h-df.de:group:HDF#login.helmholtz-data-federation.de"
        ],
        "eduperson_principal_name": "lo0018@kit.edu",
        "eduperson_scoped_affiliation": "member@kit.edu",
        "eduperson_unique_id": "6c611e2a2c1c487f9948c058a36c8f0e@login.helmholtz-data-federation.de",
        "email": "marcus.hardt@kit.edu",
        "email_verified": "true",
        "family_name": "Hardt",
        "given_name": "Marcus",
        "groups": [
            "/wlcg-test",
            "/IMK-TRO-EWCC",
            "/MyExampleColab",
            "/HDF",
            "/"
        ],
        "name": "Marcus Hardt",
        "preferred_username": "marcus",
        "sn": "Hardt",
        "ssh_key": "ssh-ed25519 AAAAC3NzaC1lZDI1NTE5AAAAIAqA5FW6m3FbFhCOsRQBxKMRki5qJxoNhZdaeLXg6ym/ marcus@nemo2019\n",
        "sub": "6c611e2a-2c1c-487f-9948-c058a36c8f0e",
        "iss": "https://login.helmholtz-data-federation.de/oauth2"
    })

INPUT_EGI = wrap_userinfo({
        "acr": "https://aai.egi.eu/LoA#Substantial",
        "eduperson_assurance": [
            "https://aai.egi.eu/LoA#Substantial"
        ],
        "email": "marcus.hardt@kit.edu",
        "family_name": "Hardt",
        "given_name": "Marcus",
        "preferred_username": "mhardt",
        "sub": "d7a53cbe3e966c53ac64fde7355956560282158ecac8f3d2c770b474862f4756@egi.eu",
        "iss": "https://aai.egi.eu/oidc/"
    })

INPUT_DEEP_IAM = wrap_userinfo({
        "iss": "https://iam.deep-hybrid-datacloud.eu/",
        "family_name": "Hardt",
        "given_name": "Marcus",
        "groups": [
            "KIT-Cloud"
        ],
        "name": "Marcus Hardt",
        "organisation_name": "deep-hdc",
        "preferred_username": "marcus",
        "sub": "d9730f60-3b19-4f45-83ab-f29addf72d58",
        "updated_at": "Mon Jun 25 16:55:15 CEST 2018"
    })

INPUT_INDIGO_IAM = wrap_userinfo({
        "iss": "https://iam-test.indigo-datacloud.eu/",
        "family_name": "Hardt",
        "gender": "M",
        "given_name": "Marcus",
        "groups": [
            "Users",
            "Developers",
            "test.vo-users"
        ],
        "name": "Marcus Hardt",
        "organisation_name": "indigo-dc",
        "preferred_username": "marcus",
        "sub": "a1ea3aa2-8daf-41bb-b4fb-eb88f439e446",
        "updated_at": 1563283972
    })

INPUT_KIT = wrap_userinfo({
        "displayName": "Hardt, Marcus (SCC)",
        "eduPersonEntitlement": [
            "urn:geant:kit.edu:group:DFN-SLCS",
            "urn:geant:kit.edu:group:LSDF-DIS",
            "urn:geant:kit.edu:group:bwGrid",
            "urn:geant:kit.edu:group:bwLSDF-FS",
            "urn:geant:kit.edu:group:bwUniCluster",
            "urn:geant:kit.edu:group:bwsyncnshare",
            "urn:geant:kit.edu:group:bwsyncnshare-idm",
            "urn:geant:kit.edu:group:gruppenverwalter"
        ],
        "eduPersonPrincipalName": "lo0018@kit.edu",
        "eduPersonScopedAffiliation": [
            "employee@kit.edu",
            "member@kit.edu"
        ],
        "eduperson_entitlement": [
            "urn:geant:kit.edu:group:DFN-SLCS",
            "urn:geant:kit.edu:group:LSDF-DIS",
            "urn:geant:kit.edu:group:bwGrid",
            "urn:geant:kit.edu:group:bwLSDF-FS",
            "urn:geant:kit.edu:group:bwUniCluster",
            "urn:geant:kit.edu:group:bwsyncnshare",
            "urn:geant:kit.edu:group:bwsyncnshare-idm",
            "urn:geant:kit.edu:group:gruppenverwalter"
        ],
        "eduperson_principal_name": "lo0018@kit.edu",
        "eduperson_scoped_affiliation": [
            "employee@kit.edu",
            "member@kit.edu"
        ],
        "email": "marcus.hardt@kit.edu",
        "family_name": "Hardt",
        "givenName": "Marcus",
        "given_name": "Marcus",
        "mail": "marcus.hardt@kit.edu",
        "name": "Marcus Hardt",
        "ou": "SCC",
        "preferred_username": "lo0018",
        "sn": "Hardt",
        "sub": "4cbcd471-1f51-4e54-97b8-2dd5177e25ec",
        "iss": "https://oidc.scc.kit.edu/auth/realms/kit/"
    })


ALL_INPUT = [
    INPUT_UNITY,
    INPUT_EGI,
    INPUT_DEEP_IAM,
    INPUT_INDIGO_IAM,
    INPUT_KIT
]


"""LDAP backend for pre-created accounts.
It"s in the proof-of-concept state.
"""


import logging
from ldap3 import (
    Server,
    Connection,
    ALL,
    BASE,
    MODIFY_REPLACE,
    MODIFY_DELETE,
    MODIFY_ADD,
    SAFE_SYNC,
)
from enum import Enum, auto

from ..config import CONFIG
from ..results import Failure, Rejection


logger = logging.getLogger(__name__)


class Mode(Enum):
    READ_ONLY = auto()
    PRE_CREATED = auto()
    FULL_ACCESS = auto()

    @staticmethod
    def from_str(label):
        label = label.lower()
        if label in ("read_only", "read-only", "readonly"):
            return Mode.READ_ONLY
        elif label in ("pre_created", "pre-created", "precreated"):
            return Mode.PRE_CREATED
        elif label in ("full_access", "full-access", "fullaccess"):
            return Mode.FULL_ACCESS
        else:
            msg = (
                f"Unknown mode '{label}'. Supported modes: "
                f"{[name for name, member in Mode.__members__.items()]}."
            )
            logger.error(msg)
            raise Failure(message=msg)


DEFAULT_MODE = Mode.READ_ONLY
DEFAULT_HOST = "localhost"
DEFAULT_PORT = 1389
DEFAULT_TLS_PORT = 636
DEFAULT_ANONYMOUS = True
DEFAULT_ADMIN_USER = None
DEFAULT_ADMIN_PASSWORD = None
DEFAULT_TLS = False
DEFAULT_USER_BASE = "ou=users,dc=example"
DEFAULT_GROUP_BASE = "ou=groups,dc=example"
DEFAULT_ATTR_OIDC_UID = "gecos"
DEFAULT_ATTR_LOCAL_UID = "uid"
DEFAULT_SHELL = "/bin/sh"
DEFAULT_HOME_BASE = "/home"
DEFAULT_UID_MIN = 1000
DEFAULT_UID_MAX = 60000
DEFAULT_GID_MIN = 1000
DEFAULT_GID_MAX = 60000


class LdapSearchResult:
    def __init__(self, ldap_connection, args, kwargs):
        """Perform a search with given arguments and create LdapSearchResult object.

        Initialise fields from the return value of search (connection in SAFE_SYNC mode),
        where the return value is a tuple (status, result, response, request).

        @param ldap_connection: an active ldap3.Connection object
        @param args: list of arguments for ldap3.Connection.search method
        @param kwargs: dictionary of key-value arguments for ldap3.Connection.search method
        """
        try:
            search_result = ldap_connection.search(*args, **kwargs)
            self.status = search_result[0]
            self.result = search_result[1]
            self.response = search_result[2]
            self.request = search_result[3]
        except Exception as e:
            msg = "Error searching in LDAP"
            logger.error(f"{msg}: {e}")
            raise Failure(message=msg)

    def found(self):
        return self.status

    def get_attribute(self, attribute_name):
        try:
            value = self.response[0]["attributes"][attribute_name]
            if isinstance(value, list):
                return value[0]
            else:
                return value
        except Exception as e:
            logger.warning(f"Attribute {attribute_name} not found in response: {e}")
            return None


class LdapConnection:
    """Connection to the LDAP server."""

    def __init__(
        self,
        mode=DEFAULT_MODE,
        host=DEFAULT_HOST,
        port=DEFAULT_PORT,
        tls=DEFAULT_TLS,
        admin_user=DEFAULT_ADMIN_USER,
        admin_password=DEFAULT_ADMIN_PASSWORD,
        user_base=DEFAULT_USER_BASE,
        group_base=DEFAULT_GROUP_BASE,
        attr_oidc_uid=DEFAULT_ATTR_OIDC_UID,
        attr_local_uid=DEFAULT_ATTR_LOCAL_UID,
        shell=DEFAULT_SHELL,
        home_base=DEFAULT_HOME_BASE,
        uid_min=DEFAULT_UID_MIN,
        uid_max=DEFAULT_UID_MAX,
        gid_min=DEFAULT_GID_MIN,
        gid_max=DEFAULT_GID_MAX,
    ):
        """Initialise connection to LDAP server.

        :param str host: host where LDAP server is running, default localhost
        :param int port: port where LDAP server is running, default 1389
        :param bool tls: whether connection to LDAP is SSL encrypted
        :param str admin_user: admin username
        :param str admin_password: admin password
        :param str user_base: base used for user namespace in LDAP operations
        :param str group_base: base used for group namespace in LDAP operations
        :param str attr_oidc_uid: LDAP attribute to store uid for federated user, default uid
        :param str attr_local_uid: LDAP attribute to store uid for local user, default gecos
        :param str shell: shell used when creating users
        :param str home_base: base directory for users' home directories
                         local username will be appended to this to create homedir
        """
        self.mode = mode
        self.user_base = user_base
        self.group_base = group_base
        self.attr_oidc_uid = attr_oidc_uid
        self.attr_local_uid = attr_local_uid
        self.shell = shell
        self.home_base = home_base
        self.uid_min = uid_min
        self.uid_max = uid_max
        self.gid_min = gid_min
        self.gid_max = gid_max

        # initialise and bind connection to LDAP
        try:
            server = Server(f"ldap://{host}:{port}", get_info=ALL)
            if admin_user and admin_password:
                # add SAFE_SYNC, so we get more return values
                self.connection = Connection(
                    server,
                    admin_user,
                    admin_password,
                    auto_bind=True,
                    client_strategy=SAFE_SYNC,
                )
            else:
                self.connection = Connection(server, auto_bind=True, client_strategy=SAFE_SYNC)
        except Exception as e:
            msg = f"Could not connect to server ldap://{host}:{port}/"
            logger.error(f"{msg}: {e}")
            raise Failure(message=msg)

    def init_nextuidgid(self):
        """Initialise uidNext and gidNext entries in FULL_ACCESS mode
        with values starting in configured range.
        """
        try:
            if self.mode == Mode.FULL_ACCESS:
                search_uid = self.search_next_uid()
                if not search_uid.found():
                    self.connection.add(
                        f"cn=uidNext,{self.user_base}",
                        object_class=["uidNext"],
                        attributes={"cn": "uidNext", "uidNumber": self.uid_min},
                    )
                else:
                    logger.info("Using existing uidNext value.")

                search_gid = self.search_next_gid()
                if not search_gid.found():
                    self.connection.add(
                        f"cn=gidNext,{self.group_base}",
                        object_class=["gidNext"],
                        attributes={"cn": "gidNext", "gidNumber": self.gid_min},
                    )
                else:
                    logger.info("Using existing gidNext value.")
        except Exception as e:
            msg = "Error adding entries in LDAP for tracking available UID and GID values"
            logger.error(f"{msg}: {e}")
            raise Failure(message=msg)

    def search_user_by_oidc_uid(self, oidc_uid):
        return LdapSearchResult(
            self.connection,
            [
                f"{self.user_base}",
                f"(&({self.attr_oidc_uid}={oidc_uid})(objectClass=inetOrgPerson)(objectClass=posixAccount))",
            ],
            {"attributes": [self.attr_local_uid]},
        )

    def search_user_by_local_username(self, username):
        return LdapSearchResult(
            self.connection,
            [
                f"{self.user_base}",
                f"(&({self.attr_local_uid}={username})(objectClass=inetOrgPerson)(objectClass=posixAccount))",
            ],
            {"attributes": [self.attr_oidc_uid]},
        )

    def search_group_by_name(self, group_name):
        return LdapSearchResult(
            self.connection,
            [
                f"{self.group_base}",
                f"(&(cn={group_name})(objectClass=posixGroup))",
            ],
            {"attributes": ["memberUid", "gidNumber"]},
        )

    def search_next_uid(self):
        return LdapSearchResult(
            self.connection,
            [
                f"{self.user_base}",
                "(&(cn=uidNext)(objectClass=uidNext))",
            ],
            {"attributes": ["uidNumber"]},
        )

    def search_next_gid(self):
        return LdapSearchResult(
            self.connection,
            [
                f"{self.group_base}",
                "(&(cn=gidNext)(objectClass=gidNext))",
            ],
            {"attributes": ["gidNumber"]},
        )

    def is_uid_taken(self, uid):
        return LdapSearchResult(
            self.connection,
            [
                f"{self.user_base}",
                f"(&(uidNumber={uid})(objectClass=posixAccount))",
            ],
            {},
        ).found()

    def is_gid_taken(self, gid):
        return LdapSearchResult(
            self.connection,
            [
                f"{self.group_base}",
                f"(&(gidNumber={gid})(objectClass=posixGroup))",
            ],
            {},
        ).found()

    def get_next_uid(self):
        search_result = self.search_next_uid()
        if search_result.found():
            uid = search_result.get_attribute("uidNumber")

            # make sure uid is not taken already
            next_uid = uid
            while self.is_uid_taken(next_uid):
                next_uid += 1

            # make sure uid still in allowed range
            if next_uid > self.uid_max:
                raise Exception("No available UIDs left in configured range.")

            # specify uid in MODIFY_DELETE operation to avoid race conditions
            # the operation will fail if the value has been modified in the meantime
            result = self.connection.modify(
                f"cn=uidNext,{self.user_base}",
                {"uidNumber": [(MODIFY_DELETE, [uid]), (MODIFY_ADD, [next_uid + 1])]},
            )
            return next_uid

    def get_next_gid(self):
        search_result = self.search_next_gid()
        if search_result.found():
            gid = search_result.get_attribute("gidNumber")

            # make sure gid is not taken already
            next_gid = gid
            while self.is_gid_taken(next_gid):
                next_gid += 1

            # make sure gid still in allowed range
            if next_gid > self.gid_max:
                raise Exception("No available GIDs left in configured range.")

            # specify gid in MODIFY_DELETE operation to avoid race conditions
            # the operation will fail if the value has been modified in the meantime
            result = self.connection.modify(
                f"cn=gidNext,{self.group_base}",
                {"gidNumber": [(MODIFY_DELETE, [gid]), (MODIFY_ADD, [next_gid + 1])]},
            )
            return next_gid

    def add_user(self, userinfo, local_username, primary_group_name):
        """Add an LDAP entry for `local_username` with
        all information from `userinfo`.
        If user exists, a Failure exception is raised.
        """
        try:
            return self.connection.add(
                f"uid={local_username},{self.user_base}",
                object_class=["top", "inetOrgPerson", "posixAccount"],
                attributes={
                    "sn": userinfo.family_name,
                    "givenName": userinfo.given_name,
                    "cn": userinfo.full_name,
                    "mail": userinfo.email,
                    "uid": local_username,
                    "uidNumber": self.get_next_uid(),
                    "gidNumber": self.search_group_by_name(primary_group_name).get_attribute(
                        "gidNumber"
                    ),
                    "homeDirectory": f"{self.home_base}/{local_username}",
                    "loginShell": self.shell,
                    self.attr_local_uid: local_username,
                    self.attr_oidc_uid: userinfo.unique_id,
                },
            )
        except Exception as e:
            msg = f"Failed to add an LDAP entry for uid {userinfo.unique_id} with local username {local_username}."
            logger.error(f"{msg}: {e}")
            raise Failure(message=msg)

    def map_user(self, userinfo, local_username):
        """Update the LDAP entry for given `local_username` with
        mapped oidc uid.
        If user doesn't exist, a Failure exception is raised.
        """
        try:
            return self.connection.modify(
                f"uid={local_username},{self.user_base}",
                {self.attr_oidc_uid: [(MODIFY_REPLACE, [userinfo.unique_id])]},
            )
        except Exception as e:
            msg = f"Failed to modify the LDAP entry for uid {userinfo.unique_id} with local username {local_username}."
            logger.error(f"{msg}: {e}")
            raise Failure(message=msg)

    def update_user(self, userinfo, local_username):
        """Update the LDAP entry for given `local_username` with
        all information in `userinfo`.
        If user doesn't exist, a Failure exception is raised.
        """
        try:
            return self.connection.modify(
                f"uid={local_username},{self.user_base}",
                {
                    "sn": [(MODIFY_REPLACE, [userinfo.family_name])],
                    "givenName": [(MODIFY_REPLACE, [userinfo.given_name])],
                    "cn": [(MODIFY_REPLACE, [userinfo.full_name])],
                    "mail": [(MODIFY_REPLACE, [userinfo.email])],
                    "uid": [(MODIFY_REPLACE, [local_username])],
                    "homeDirectory": [(MODIFY_REPLACE, [f"{self.home_base}/{local_username}"])],
                    "loginShell": [(MODIFY_REPLACE, [self.shell])],
                    self.attr_local_uid: [(MODIFY_REPLACE, [local_username])],
                    self.attr_oidc_uid: [(MODIFY_REPLACE, [userinfo.unique_id])],
                },
            )
        except Exception as e:
            msg = f"Failed to modify the LDAP entry for uid {userinfo.unique_id} with local username {local_username}."
            logger.error(f"{msg}: {e}")
            raise Failure(message=msg)

    def delete_user(self, local_username):
        """Delete the LDAP entry for given `local_username`.
        If user doesn't exist, a Failure exception is raised.
        """
        try:
            return self.connection.delete(f"uid={local_username},{self.user_base}")
        except Exception as e:
            msg = f"Failed to delete the LDAP entry for local username {local_username}."
            logger.error(f"{msg}: {e}")
            raise Failure(message=msg)

    def add_user_to_group(self, local_username, group_name):
        """Add a user to group.
        If either of them does not exist, a Failure exception is raised.
        """
        try:
            self.connection.modify(
                f"cn={group_name},{self.group_base}",
                {
                    "memberUid": [(MODIFY_ADD, [local_username])],
                },
            )
        except Exception as e:
            msg = f"Failed to modify the LDAP entry for group {group_name} with local username {local_username}."
            logger.error(f"{msg}: {e}")
            raise Failure(message=msg)

    def add_group(self, group_name):
        """Add an LDAP entry for `group_name`.
        If group exists, a warning is issued.
        """
        try:
            return self.connection.add(
                f"cn={group_name},{self.group_base}",
                object_class=["top", "posixGroup"],
                attributes={
                    "cn": group_name,
                    "gidNumber": self.get_next_gid(),
                },
            )
        except Exception as e:
            msg = f"Failed to add an LDAP entry for group {group_name}."
            logger.error(f"{msg}: {e}")
            raise Failure(message=msg)

    @staticmethod
    def load():
        try:
            config = CONFIG["backend.ldap"]
            mode = Mode.from_str(config.get("mode", DEFAULT_MODE))
            host = config.get("host", DEFAULT_HOST)
            tls = config.get("tls", DEFAULT_TLS)
            if tls:
                port = config.getint("port", DEFAULT_TLS_PORT)
            else:
                port = config.getint("port", DEFAULT_PORT)
            admin_user = config.get("admin_user", DEFAULT_ADMIN_USER)
            admin_password = config.get("admin_password", DEFAULT_ADMIN_PASSWORD)
            user_base = config.get("user_base", DEFAULT_USER_BASE)
            group_base = config.get("group_base", DEFAULT_GROUP_BASE)
            attr_oidc_uid = config.get("attribute_oidc_uid", DEFAULT_ATTR_OIDC_UID)
            attr_local_uid = config.get("attribute_local_uid", DEFAULT_ATTR_LOCAL_UID)

            # only needed in FULL_ACCESS mode
            shell = config.get("shell", DEFAULT_SHELL)
            home_base = config.get("home_base", DEFAULT_HOME_BASE).rstrip("/")
            uid_min = config.getint("uid_min", DEFAULT_UID_MIN)
            uid_max = config.getint("uid_max", DEFAULT_UID_MAX)
            gid_min = config.getint("gid_min", DEFAULT_GID_MIN)
            gid_max = config.getint("gid_max", DEFAULT_GID_MAX)

            ldap = LdapConnection(
                mode,
                host,
                port,
                tls,
                admin_user,
                admin_password,
                user_base,
                group_base,
                attr_oidc_uid,
                attr_local_uid,
                shell,
                home_base,
                uid_min,
                uid_max,
                gid_min,
                gid_max,
            )
        except KeyError:
            logger.warning(
                "Could not find [backend.ldap] section in feudal config, using defaults..."
            )
            ldap = LdapConnection()
        # init uidNext and gidNext entries in LDAP
        ldap.init_nextuidgid()
        return ldap


LDAP = LdapConnection.load()


class User:
    """Manages the user object on the service."""

    def __init__(self, userinfo):
        """
        Arguments:
        userinfo -- (type: UserInfo)
        """
        self.primary_group = Group(userinfo.primary_group)
        self.userinfo = userinfo
        self.unique_id = userinfo.unique_id
        logger.debug(f"backend processing: {userinfo.unique_id}")
        if self.exists():
            username = self.get_username()
            logger.debug(f"This user does actually exist. The name is: {username}")
            self.set_username(username)
        else:
            self.set_username(userinfo.username)

        self.ssh_keys = [key["value"] for key in userinfo.ssh_keys]
        self.credentials = {}

    def exists(self):
        """Return whether the user exists on the service.

        If this returns True,  calling `create` should have no effect or raise an error.
        """
        logger.info(f"Check if user exists: {self.unique_id}")
        return LDAP.search_user_by_oidc_uid(self.unique_id).found()

    def name_taken(self, name):
        """Return whether the username is already taken by another user on the service,
        i.e. if an entry for it exists in the LDAP and it's not mapped to the current oidc uid.

        In 'pre_created' mode, taken means that the username has been mapped to another oidc uid.
        """
        search_result = LDAP.search_user_by_local_username(name)
        if search_result.found():  # there is an entry for name in LDAP
            oidc_uid = search_result.get_attribute(
                LDAP.attr_oidc_uid
            )  # this is the oidc_uid mapped to name
            if LDAP.mode == Mode.PRE_CREATED and oidc_uid is None:  # pre-created but not mapped
                return False
            return oidc_uid != self.unique_id  # name already mapped to another oidc uid
        else:  # no entry for name found in LDAP
            return False

    def get_username(self):
        """Check if a user exists based on unique_id and return the name"""
        return LDAP.search_user_by_oidc_uid(self.unique_id).get_attribute(LDAP.attr_local_uid)

    def set_username(self, username):
        """Set local username on the service."""
        self.name = username

    def create(self):
        """Create the user on the service.

        If the user already exists, do nothing or raise an error
        """
        if LDAP.mode == Mode.READ_ONLY:
            msg = (
                f"LDAP backend in read_only mode, new entry cannot be added for user {self.unique_id}."
                f" (local username {self.name})"
            )
            logger.error(msg)
            raise Rejection(
                message=f"{msg} Please contact an administrator to create an account for you."
            )
        elif LDAP.mode == Mode.PRE_CREATED:
            if not LDAP.search_user_by_local_username(self.name).found():
                msg = f"Local username {self.name} not found in LDAP for user {self.unique_id}."
                logger.error(msg)
                raise Failure(
                    message=f"{msg} Please contact an administrator to pre-create this account for you."
                )
            else:
                LDAP.map_user(self.userinfo, self.name)
        else:  # Mode.FULL_ACCESS
            LDAP.add_user(self.userinfo, self.name, self.primary_group.name)

    def update(self):
        """Update all relevant information about the user on the service.

        If the user doesn't exists, behaviour is undefined.
        """
        self.credentials["ssh_user"] = self.name
        self.credentials["ssh_host"] = CONFIG["backend.ldap.login_info"].get(
            "ssh_host", "undefined"
        )
        self.credentials["commandline"] = "ssh {}@{}".format(
            self.credentials["ssh_user"], self.credentials["ssh_host"]
        )

        if LDAP.mode == Mode.READ_ONLY:
            msg = (
                f"LDAP backend in read_only mode, entry for user {self.unique_id} "
                f"cannot be modified."
            )
            logger.warning(msg)
        elif LDAP.mode == Mode.PRE_CREATED:
            msg = (
                f"LDAP backend in pre_created mode, entry for user {self.unique_id} "
                f"cannot be modified."
            )
            logger.warning(msg)
        else:  # Mode.FULL_ACCESS
            LDAP.update_user(self.userinfo, self.name)

    def delete(self):
        """Delete the user on the service.

        If the user doesn"t exists, do nothing or raise an error.
        """
        if LDAP.mode == Mode.READ_ONLY:
            msg = f"LDAP backend in read_only mode, entry for local username {self.name} cannot be deleted."
            logger.error(msg)
            raise Rejection(message=msg)
        elif LDAP.mode == Mode.PRE_CREATED:
            msg = f"LDAP backend in pre_created mode, entry for local username {self.name} cannot be deleted."
            logger.error(msg)
            raise Failure(message=msg)
        else:  # Mode.FULL_ACCESS
            LDAP.delete_user(self.name)

    def mod(self, supplementary_groups=None):
        """Modify the user on the service.

        The state of the user with respect to the provided Arguments after calling this function
        should not depend on the state the user had previously.

        If the user doesn't exists, behaviour is undefined.

        Arguments:
        supplementary_groups -- A list of groups to add the user to (type: list(Group))
        """
        if supplementary_groups is None or supplementary_groups == []:
            logger.debug(f"Empty group list for user {self.name}. Nothing to do here.")
        else:
            logger.debug(
                f"Ensuring user '{self.name}' is member of these groups: \
                         {[g.name for g in supplementary_groups]}"
            )
            if LDAP.mode == Mode.READ_ONLY:
                msg = f"LDAP backend in read_only mode, local username {self.name} cannot be added to given groups."
                logger.warning(msg)
            elif LDAP.mode == Mode.PRE_CREATED:
                for group in supplementary_groups:
                    LDAP.add_user_to_group(self.name, group.name)
            else:  # Mode.FULL_ACCESS
                for group in supplementary_groups:
                    LDAP.add_user_to_group(self.name, group.name)

    def install_ssh_keys(self):
        """Install users SSH keys on the service.

        No other SSH keys should be active after calling this function.

        If the user doesn't exists, behaviour is undefined.
        """
        # TODO: use ldapPublicKey schema (sshPublicKey attribute) for storing ssh key in LDAP
        pass

    def uninstall_ssh_keys(self):
        """Uninstall the users SSH keys on the service.

        This must uninstall all SSH keys installed with `install_ssh_keys`. It may uninstall SSH
        keys installed by other means.

        If the user doesn't exists, behaviour is undefined.
        """
        # TODO: use ldapPublicKey schema (sshPublicKey attribute) for storing ssh key in LDAP
        pass


class Group:
    """Manages the group object on the service."""

    def __init__(self, name):
        """
        Arguments:
        name -- The name of the group
        """
        self.name = name

    def exists(self):
        """Return whether the group already exists."""
        logger.debug(f"Check if group exists: {self.name}")
        if LDAP.search_group_by_name(self.name).found():
            logger.debug(f"Group {self.name} exists.")
            return True
        else:
            logger.debug(f"Group {self.name} doesn't exist.")
            return False

    def create(self):
        """Create the group on the service.

        If the group already exists, nothing happens.
        """
        if self.exists():
            logger.info(f"Group {self.name} exists.")
        elif LDAP.mode == Mode.READ_ONLY:
            msg = (
                f"LDAP backend in read_only mode, new entry cannot be added for group {self.name}."
            )
            logger.warning(msg)
        elif LDAP.mode == Mode.PRE_CREATED:
            msg = f"LDAP backend in pre_created mode, new entry cannot be added for group {self.name}."
            logger.warning(msg)
        else:  # Mode.FULL_ACCESS
            LDAP.add_group(self.name)



import pytest
import regex
import random
from pathlib import Path

from ldf_adapter.backend.local_unix import make_shadow_compatible
from ldf_adapter.results import Failure


INPUT_UNIX = {
    "new_root": f"/tmp/newroot{random.randint(1000, 9999)}",
    "userinfo": {
        "unique_id": "subuid@issuer.domain",
        "primary_group": "testgroup",
        "username": "testuser",
        "ssh_keys": {}
    },
    "passwd_entry": "testuser:x:1000:1000:subuid@issuer.domain:/home/testuser:/bin/bash",
    "passwd_taken": "testuser:x:2000:2000:::",
    "group_entry": "testgroup:x:1000:",
}


@pytest.mark.parametrize('input,exists,taken', [(INPUT_UNIX, False, False)])
def test_create(local_unix_user, input):
    """Test that create method adds the appropriate entry in /etc/passwd.
    """
    local_unix_user.create()
    assert input["passwd_entry"] in (Path(input["new_root"])/"etc"/"passwd").read_text()


@pytest.mark.parametrize('input,exists,taken', [(INPUT_UNIX, False, True)])
def test_create_taken(local_unix_user, input):
    """Test that create raises a Failure if the username is already taken.
    User's primary group needs to exist.
    """
    with pytest.raises(Failure):
        local_unix_user.create()


@pytest.mark.parametrize('input,exists,taken', [
        (INPUT_UNIX, False, False),
        (INPUT_UNIX, False, True)
    ])
def test_doesnt_exist(local_unix_user):
    """Tests that exists returns False on a clean system, or even if username is taken by another user.
    (i.e. username exists, but the gecos field does not contain this user's unique_id)."""
    assert not local_unix_user.exists()


@pytest.mark.parametrize('input,exists,taken', [(INPUT_UNIX, True, False)])
def test_exists(local_unix_user, input):
    """Tests that exists returns True if there is an entry in passwd for this unique_id."""
    assert local_unix_user.exists()


@pytest.mark.parametrize('input,exists,taken', [(INPUT_UNIX, False, True)])
def test_name_taken(local_unix_user, input):
    """Tests that name_taken returns True if the username exists on the system but is not
    mapped to this user's unique_id (via gecos field).
    """
    assert local_unix_user.name_taken(input["userinfo"]["username"])


@pytest.mark.parametrize('input,exists,taken', [(INPUT_UNIX, True, False)])
def test_name_taken_byme(local_unix_user, input):
    """Tests that name_taken returns False if the username exists on the system and is
    mapped to this user's unique_id (via gecos field).
    """
    assert not local_unix_user.name_taken(input["userinfo"]["username"])


@pytest.mark.parametrize('input,exists,taken', [(INPUT_UNIX, True, False)])
def test_get_username_same(local_unix_user, input):
    """Tests that get_username returns the username in passwd.
    Case 1: username in userinfo is the same as the one in passwd.
    """
    assert local_unix_user.get_username() == "testuser"


@pytest.mark.parametrize('input,exists,taken', [(INPUT_UNIX, False, False)])
def test_get_username_different(local_unix_user, input):
    """Tests that get_username returns the username in passwd.
    Case 2: username in userinfo is different than the one in passwd.
    """
    passwd_mapped = "testuser2:x:1000:1000:subuid@issuer.domain:/home/testuser:/bin/bash"
    (Path(input["new_root"])/"etc"/"passwd").write_text(passwd_mapped)
    assert local_unix_user.get_username() == "testuser2"


@pytest.mark.parametrize('input,exists,taken', [
        (INPUT_UNIX, False, False),
        (INPUT_UNIX, False, True)
    ])
def test_get_username_none(local_unix_user, input):
    """Tests that get_username returns None if there is no username mapped to the
    user's unique_id, even when the username in userinfo is already taken.
    """
    assert local_unix_user.get_username() == None


@pytest.mark.parametrize('input,exists,taken', [(INPUT_UNIX, True, False)])
def test_delete(local_unix_user, input):
    """Test that after delete, there is no user with the username in the db,
    or any other user mapped to the unique_id."""
    local_unix_user.delete()
    passwd_content = (Path(input["new_root"])/"etc"/"passwd").read_text()
    assert input["userinfo"]["unique_id"] not in passwd_content
    assert input["userinfo"]["username"] not in passwd_content


@pytest.mark.parametrize('input,exists,taken', [
        (INPUT_UNIX, False, False),
        (INPUT_UNIX, False, True)
    ])
def test_delete_doesnt_exist(local_unix_user, input, taken):
    """Test that delete raises a Failure if the user is not in the db,
    even if its username is taken by another user."""
    with pytest.raises(Failure):
        local_unix_user.delete()
    passwd_content = (Path(input["new_root"])/"etc"/"passwd").read_text()
    assert input["userinfo"]["unique_id"] not in passwd_content
    if taken:
        assert input["userinfo"]["username"] in passwd_content


INPUT_UNIX_GROUP = {
    "new_root": f"/tmp/newroot{random.randint(1000, 9999)}",
    "name": "testgroup",
    "group_entry": "testgroup:x:1000:",
}


@pytest.mark.parametrize('input,exists', [(INPUT_UNIX_GROUP, False)])
def test_group_create(local_unix_group, input):
    """Test that create method adds the appropriate entry in /etc/group.
    """
    local_unix_group.create()
    assert input["group_entry"] in (Path(input["new_root"])/"etc"/"group").read_text()


@pytest.mark.parametrize('input,exists', [(INPUT_UNIX_GROUP, True)])
def test_group_create_exists(local_unix_group):
    """Test that create raises a Failure if the group exists.
    """
    with pytest.raises(Failure):
        local_unix_group.create()


@pytest.mark.parametrize('input,exists', [
        (INPUT_UNIX_GROUP, False),
        (INPUT_UNIX_GROUP, True)
    ])
def test_group_exist(local_unix_group, exists):
    """Tests that exists returns True if there is an entry for the given name,
    and False otherwise.
    """
    assert local_unix_group.exists() == exists



INPUT_SHADOW_COMPATIBLE = [
    ("user", "user"),
    ("", "_"),
    ("äöüÄÖÜß!$*@", "aeoeueaeoeuessisx_at_"),
    ("#%^&()=+[]{}\\|;:'\",<.>/?", "________________________"),
    ("u#%^&()=+[]{}\\|;:'\",<.>/?", "u________________________"),
    (u"\u5317\u4EB0", "bei_jing_"),
    (u"\u20AC", "eur"),
    ("user$", "users"),
    ("-user", "_-user"),
    ("-aaaaaaaaaaaaaaaaaaaaaaaaaaaaaa", "_-aaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"),
    ("-aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa", "_aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"),
    ("-abcdefaaaaaaaaaaaaaaaaaaaaaaaaaa", "_..defaaaaaaaaaaaaaaaaaaaaaaaaaa"),
    ("abcdefaaaaaaaaaaaaaaaaaaaaaaaaaaa", "__defaaaaaaaaaaaaaaaaaaaaaaaaaaa"),
    ("helmholtz-de_KIT_Helmholtz-member", "helmholtz-de_.._helmholtz-member")
    # ("a_b_c_d_e_f_a_a_a_a______________", "a_.._d_e_f_a_a_a_a______________"), # ??
    # ("_________________________________", "_.._____________________________"), # ??
]


@pytest.mark.parametrize('raw', [x[0] for x in INPUT_SHADOW_COMPATIBLE])
def test_make_shadow_compatible_length(raw):
    """a shadow-compatible name must be at most 32 characters long
    """
    assert len(make_shadow_compatible(raw)) <= 32


@pytest.mark.parametrize('raw', [x[0] for x in INPUT_SHADOW_COMPATIBLE])
def test_make_shadow_compatible_allowed_chars(raw):
    """a shadow-compatible name must start with a lowercase letter or underscore
    and can also contain numbers and - in addition to lowercase letters and _
    """
    word = make_shadow_compatible(raw)
    assert regex.match(r'[a-z_]', word[0]) and regex.match(r'[-0-9_a-z]', word)


@pytest.mark.parametrize('raw,cooked', INPUT_SHADOW_COMPATIBLE)
def test_make_shadow_compatible(raw, cooked):
    """expected behaviour:
    - german umlauts are replaced with their phonetic equivalents
    - a few special characters are replaced by sensible equivalents:
        - ! to i
        - $ to s
        - * to x
        - @ to _at_
    - unicode characters are decoded to ascii
    - all other special characters are replaced with _
    - what about shortening? TODO: define expected behaviour
    """
    assert make_shadow_compatible(raw) == cooked
